/* override table width restrictions */
@media screen and (min-width: 767px) {
  .wy-table-responsive table td {
    /* !important prevents the common CSS stylesheets from overriding
         this as on RTD they are loaded after this stylesheet */
    white-space: normal !important;
  }

  .wy-table-responsive {
    overflow: visible !important;
  }
}

div.admonition.caution {
  background: #ffc3c3;
}
.rst-content .caution .admonition-title {
  background: #ff6f6f;
}

:root {
  --pst-color-primary: #5e409d;
  --pst-color-link: #5e409d;
  --pst-color-link-hover: #483271;
  --bs-link-color: #5e409d;
  --bs-link-hover-color: #483271;
  --pst-color-inline-code-links: #24837b;

  /* Add dark mode colors */
  --pst-color-primary-dark: #8b7ec8;
  --pst-color-link-dark: #8b7ec8;
  --pst-color-link-hover-dark: #6e59a8;
  --bs-link-color-dark: #8b7ec8;
  --bs-link-hover-color-dark: #6e59a8;
  --pst-color-inline-code-links-dark: #3aa99f;
}

/* Theme colors */
html[data-theme="light"] {
  --pst-color-primary: #5e409d;
  --pst-color-link: #5e409d;
  --pst-color-link-hover: #483271;
  --bs-link-color: #5e409d;
  --bs-link-hover-color: #483271;
  --pst-color-inline-code-links: #24837b;
}

html[data-theme="dark"] {
  --pst-color-primary: #8b7ec8;
  --pst-color-link: #8b7ec8;
  --pst-color-link-hover: #6e59a8;
  --bs-link-color: #8b7ec8;
  --bs-link-hover-color: #6e59a8;
  --pst-color-inline-code-links: #3aa99f;
}

/* Ensure styles are applied */
.bd-header {
  background-color: var(--pst-color-primary);
}

a {
  color: var(--pst-color-link);
}

a:hover {
  color: var(--pst-color-link-hover);
}
