main {
  min-width: 1000px !important;
}
.iframe-container {
  width: 100%;
  height: 600px;
  overflow: hidden;
  margin: 20px 0;
}
.iframe-container iframe {
  width: 100%;
  height: 100%;
  border: none;
}
.iframe-container div {
  border: 1px solid #ddd;
  border-radius: 4px;
}
.image-container {
  width: 100%;
  text-align: center;
  margin: 20px 0;
}
.image-container img {
  max-width: 100%;
  height: auto;
}
